/**
 * @packageDocumentation
 *
 * A Datastore implementation for browsers that stores data in [IndexedDB](https://developer.mozilla.org/en-US/docs/Web/API/IndexedDB_API).
 *
 * @example
 *
 * ```js
 * import { IDBDatastore } from 'datastore-idb'
 *
 * const store = new IDBDatastore('path/to/store')
 * ```
 */
import { BaseDatastore } from 'datastore-core';
import { openDB, deleteDB } from 'idb';
import { Key } from 'interface-datastore';
import { DeleteFailedError, GetFailedError, NotFoundError, OpenFailedError, PutFailedError } from 'interface-store';
import filter from 'it-filter';
import sort from 'it-sort';
import { raceSignal } from 'race-signal';
export class IDBDatastore extends BaseDatastore {
    location;
    version;
    db;
    constructor(location, init = {}) {
        super();
        this.location = `${init.prefix ?? ''}${location}`;
        this.version = init.version ?? 1;
    }
    async open() {
        try {
            const location = this.location;
            this.db = await openDB(location, this.version, {
                upgrade(db) {
                    db.createObjectStore(location);
                }
            });
        }
        catch (err) {
            throw new OpenFailedError(String(err));
        }
    }
    async close() {
        this.db?.close();
    }
    async put(key, val, options) {
        if (this.db == null) {
            throw new Error('Datastore needs to be opened.');
        }
        try {
            options?.signal?.throwIfAborted();
            await raceSignal(this.db.put(this.location, val, key.toString()), options?.signal);
        }
        catch (err) {
            throw new PutFailedError(String(err));
        }
        return key;
    }
    async get(key, options) {
        if (this.db == null) {
            throw new Error('Datastore needs to be opened.');
        }
        let val;
        try {
            options?.signal?.throwIfAborted();
            val = await raceSignal(this.db.get(this.location, key.toString()), options?.signal);
        }
        catch (err) {
            throw new GetFailedError(String(err));
        }
        if (val === undefined) {
            throw new NotFoundError();
        }
        return val;
    }
    async has(key, options) {
        if (this.db == null) {
            throw new Error('Datastore needs to be opened.');
        }
        try {
            options?.signal?.throwIfAborted();
            const result = await raceSignal(this.db.getKey(this.location, key.toString()));
            return Boolean(result);
        }
        catch (err) {
            throw new GetFailedError(String(err));
        }
    }
    async delete(key, options) {
        if (this.db == null) {
            throw new Error('Datastore needs to be opened.');
        }
        try {
            options?.signal?.throwIfAborted();
            await raceSignal(this.db.delete(this.location, key.toString()), options?.signal);
        }
        catch (err) {
            throw new DeleteFailedError(String(err));
        }
    }
    batch() {
        const puts = [];
        const dels = [];
        return {
            put(key, value) {
                puts.push({ key, value });
            },
            delete(key) {
                dels.push(key);
            },
            commit: async (options) => {
                if (this.db == null) {
                    throw new Error('Datastore needs to be opened.');
                }
                options?.signal?.throwIfAborted();
                const tx = this.db.transaction(this.location, 'readwrite');
                try {
                    const ops = puts.filter(({ key }) => {
                        // don't put a key we are about to delete
                        return dels.find(delKey => delKey.toString() === key.toString()) == null;
                    })
                        .map(put => {
                        return async () => {
                            await tx.store.put(put.value, put.key.toString());
                        };
                    })
                        .concat(dels.map(key => {
                        return async () => {
                            await tx.store.delete(key.toString());
                        };
                    }))
                        .concat(async () => {
                        await tx.done;
                    });
                    options?.signal?.throwIfAborted();
                    await raceSignal(Promise.all(ops.map(async (op) => { await op(); })), options?.signal);
                }
                catch {
                    tx.abort();
                }
            }
        };
    }
    async *query(q, options) {
        let it = this.#queryIt(q, (key, value) => {
            return { key, value };
        }, options);
        if (Array.isArray(q.filters)) {
            it = q.filters.reduce((it, f) => filter(it, f), it);
        }
        if (Array.isArray(q.orders)) {
            it = q.orders.reduce((it, f) => sort(it, f), it);
        }
        yield* it;
    }
    async *queryKeys(q, options) {
        let it = this.#queryIt(q, (key) => key, options);
        if (Array.isArray(q.filters)) {
            it = q.filters.reduce((it, f) => filter(it, f), it);
        }
        if (Array.isArray(q.orders)) {
            it = q.orders.reduce((it, f) => sort(it, f), it);
        }
        yield* it;
    }
    async *#queryIt(q, transform, options) {
        if (this.db == null) {
            throw new Error('Datastore needs to be opened.');
        }
        let yielded = 0;
        let index = -1;
        options?.signal?.throwIfAborted();
        for (const key of await this.db.getAllKeys(this.location)) {
            options?.signal?.throwIfAborted();
            if (q.prefix != null && !key.toString().startsWith(q.prefix)) {
                continue;
            }
            if (q.limit != null && yielded === q.limit) {
                return;
            }
            index++;
            if (q.offset != null && index < q.offset) {
                continue;
            }
            const k = new Key(key.toString());
            let value;
            try {
                value = await this.get(k, options);
            }
            catch (err) {
                if (err.name !== 'NotFoundError') {
                    throw err;
                }
                continue;
            }
            if (value == null) {
                continue;
            }
            yield transform(k, value);
            options?.signal?.throwIfAborted();
            yielded++;
        }
    }
    async destroy() {
        await deleteDB(this.location);
    }
}
//# sourceMappingURL=index.js.map